--[[----

	DeclinerAllowed is the main player check function, the bouncer, which decides if a player is a stranger or not.
	Usage, such as in another addon or a WeakAura:

			DeclinerAllowed("name-realm")

		Returns true if _you_ recently whispered "name", they're in your group, or online in your guild, friends, or bnet.
		Returns false if they are a complete stranger.
		Return is subject to settings. If Decliner is disabled, returns false. If "name" is in a list that is declined, returns false.

			DeclinerAllowed("name-realm","action")

		Returns true or false for "name" based on your settings for that action.
		Actions are "chan", "duel", "pet", "group", "guild", "chart", or "trade".

	For example, DeclinerAllowed("Guildie") will return true because Guildie is online in your guild, but also returns true if in your group.

	An example of the second usage, let's say your options will decline duel requests from guild members and realm friends, but not group members.
	Mcstabyrouge is a guildie, but not a friend, DeclinerAllowed("Mcstabyrouge","duel") will return false.
	You add them as a friend, DeclinerAllowed("Mcstabyrouge","duel") still returns false.
	You invite them to a group, DeclinerAllowed("Mcstabyrouge","duel") now returns true.

	----    If no realm is provided for "name", your realm is assumed.    ----

--------

	DeclinerQuery is similar to the above function, but ignores settings. This simply queries Decliner's memory.
	Usage, such as in another addon or a WeakAura:

			DeclinerQuery("name-realm")

		Returns true if "name" is in your group, or online in your guild, friends, or bnet.
		Returns false if they are a complete stranger.

			DeclinerQuery("name-realm","list")

		Returns true or false for "name" based on which list you're querying.
		Lists are "group", "guild", "friend", or "bnet". You can query the "whisper" list using DeclinerTrack below.

	Keep in mind Decliner only tracks online players, except for groups. Offline group members are tracked.

	----    If no realm is provided for "name", your realm is assumed.    ----

--------

	DeclinerTrack returns the amount of times you recently whispered players or they were declined.
	Players are cleared as Decliner forgets whispers or they don't spam you after some time.
	Usage, such as in another addon or a WeakAura:

			DeclinerTrack("action","name-realm")

		Returns the number of times that action occurred recently from "name".

			DeclinerTrack("action")

		Returns how many players are currently listed for that action.

	Actions are "whisper", "chan", "duel", "pet", "group", "guild", "chart", or "trade".
	"whisper" is the amount of times _you_ whispered "name" while the others are the times they were blocked for that action.

	For example, Suprawsom spammed a duel request 5 times, so DeclinerTrack("duel","Suprawsom") will return 5, then 4, and so on as time goes by.
	Once the last action is cleared, DeclinerTrack("duel","Suprawsom") will be nil, allowing Decliner to once again notify you of any further duels from them.
	Another example, you whispered Chanter 9 times due to a conversation, so DeclinerTrack("whisper","Chanter") will be 9.
	They will be authorized for whatever your options are set to for whispers, such as trading, until that number counts down to nil.

	An example of the second usage, let's say you are standing in front of Orgrimmar on the Area 52 server.
	14 players just spammed you with duel invites one right after another. DeclinerTrack("duel") will return 14.
	Depending on how many times each player spammed you, that number will count down erratically.
	It may jump to nil if all 14 only sent one request or hang around at a lower number for a while when the single requests are cleared.

	----    If no realm is provided for "name", your realm is assumed.    ----

----]]--

 -- localize lua functions

local date=date
local error=error
local format=string.format
local gmatch=string.gmatch
local gsub=string.gsub
local insert=table.insert
local ipairs=ipairs
local lower=string.lower
local match=string.match
local max=math.max
local next=next
local random=math.random
local remove=table.remove
local select=select
local sort=table.sort
local sub=string.sub
local time=time
local tonumber=tonumber
local tostring=tostring
local type=type
local upper=string.upper

 -- localize Blizzard functions

local After=C_Timer.After
local BNGetFriendGameAccountInfo=BNGetFriendGameAccountInfo
local BNGetNumFriendGameAccounts=BNGetNumFriendGameAccounts
local BNGetNumFriends=BNGetNumFriends
local CancelDuel=CancelDuel
local CancelPetDuel=C_PetBattles.CancelPVPDuel
local ClosePetition=ClosePetition
local CloseTrade=CloseTrade
local CreateFrame=CreateFrame
local ct=CopyTable
local DeclineChannelInvite=DeclineChannelInvite
local DeclineGroup=DeclineGroup
local DeclineGuild=DeclineGuild
local GetCVar=GetCVar
local GetFriendInfo=GetFriendInfo
local GetGuildRosterInfo=GetGuildRosterInfo
local GetNumFriends=GetNumFriends
local GetNumGroupMembers=GetNumGroupMembers
local GetNumGuildMembers=GetNumGuildMembers
local GetPetitionInfo=GetPetitionInfo
local HideUIPanel=HideUIPanel
local IsInRaid=IsInRaid
local PlaySoundKitID=PlaySoundKitID
local SendAddonMessage=SendAddonMessage
local SetAutoDeclineGuildInvites=SetAutoDeclineGuildInvites
local SetCVar=SetCVar
local split=strsplit
local StaticPopupSpecial_Hide=StaticPopupSpecial_Hide
local StaticPopup_Hide=StaticPopup_Hide
local StopSound=StopSound
local Ticker=C_Timer.NewTicker
local UnitName=UnitName
local ValueToBoolean=ValueToBoolean
local wipe=wipe

 -- initialize

local _
local a,L=...
local d={} -- data
local e={} -- events
local f={} -- functions
local g='|cff00ff00'
local r='|cffff0000'
local s=ChatTypeInfo.SYSTEM
local DCL=CreateFrame('frame',a)
local DCLPanel=CreateFrame('frame',a..'Panel')
DCL:RegisterEvent('ADDON_LOADED')
DCL:RegisterEvent('PLAYER_LOGIN')
DCL:RegisterEvent('PLAYER_LOGOUT')
d.file={
	ch='Interface\\RaidFrame\\ReadyCheck-Ready',
	cr='Interface\\RaidFrame\\ReadyCheck-NotReady',
	ha='Interface\\Icons\\Spell_Misc_EmotionHappy',
	sa='Interface\\Icons\\Spell_Misc_EmotionSad',
}
d.def={
	enable={chan=true,duel=true,pet=true,group=true,guild=true,chart=true,trade=true,},
	tell={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	group={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	guild={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	friend={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	bnet={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	msg={chan=false,duel=false,pet=false,group=false,guild=false,chart=false,trade=false,},
	chan={emote=true,say=true,yell=true,[1]=true,[2]=true,[26]=true,},
	gen={openinv=false,msgs=false,disable=false,spec=false,},
	open={'1','invite','portal',},
}
local dlog,l,pool={},{},{}
declinerlog,d.PLtrack,l.group,l.guild,l.friend,l.bnet={},{whisper={}},{},{},{},{}
for k in next,d.def.enable do d.PLtrack[k]={} end

 -- global string strip
 -- idea from Phanx

function f.pattern(s)
	s=gsub(s,'%%%d?$?%a','') -- removes all specifiers (%d, %s, %1$s, etc)
	s=gsub(s,'[:%(%)]','') -- removes colons and parentheses
	s=match(s,'^%s*(.-)%s*$') -- removes all space at either end of string
	return s
end

 -- Blizzard global strings used
 -- everything in caps can be found in FrameXML/GlobalStrings.lua
 -- this table is only for keeping strings in one place

d.S={
	ADD=ADD.."/"..REMOVE,
	BAT=BATTLENET_OPTIONS_LABEL,
	CAL=gsub(BN_INLINE_TOAST_ALERT,BATTLENET_OPTIONS_LABEL,CHAT),
	CHA=CHANNEL,
	CHS=CHANNELS,
	CSS=CHARACTER.." "..SETTINGS,
	DIO=DISABLE.." "..OPENING.." "..GROUP.." ",
	DIS=DISABLE.." ",
	DUE=DUEL,
	EMO=EMOTE,
	ENA=ENABLE.." "..DECLINE,
	GEN=GENERAL,
	GRM=GROUP.." "..MEMBERS,
	GRO=GROUP,
	GUI=GUILD,
	GUM=GUILD.." "..MEMBERS,
	INV=f.pattern(GUILDEVENT_TYPE_INVITE):gsub('^%a',upper),
	LOO=LFG_TITLE,
	NOS=NO.." "..SOUND,
	PET=PET_BATTLE_PVP_QUEUE,
	PTI=f.pattern(PETITION_TITLE),
	REA=lower(READY),
	REF=f.pattern(FRIENDS_LIST_REALM).." "..FRIENDS,
	REQ=GUILDINFOTAB_APPLICANTS_NONE,
	SAY=SAY,
	TRA=TRADE,
	VER=lower(GAME_VERSION_LABEL),
	WHI=f.pattern(AUTOCOMPLETE_LABEL_INTERACTED),
	WOW=BNET_CLIENT_WOW,
	YEL=YELL,
}

 -- matching tables for event, sound, and text filters

d.EV={
	CHANNEL_INVITE_REQUEST=true,
	DUEL_REQUESTED=true,
	GUILD_INVITE_REQUEST=true,
	PARTY_INVITE_REQUEST=true,
	PET_BATTLE_PVP_DUEL_REQUESTED=true,
	PETITION_SHOW=true,
	TRADE_SHOW=true,
}
d.SO={
	igCharacterInfoClose=840,
	igCharacterInfoOpen=839,
	igMainMenuClose=851,
	igMainMenuOpen=850,
	igPlayerInvite=880,
	igQuestListClose=876,
	igQuestListOpen=875,
}
d.SY={
	duel1=gsub(DUEL_REQUESTED,'%%%d?$?s','(.*)'),
	pet1=gsub(PET_BATTLE_PVP_DUEL_REQUESTED,'%%%d?$?s','(.*)'),
	guild1=gsub(gsub(ERR_INVITED_TO_GUILD_SSS,'%%%d?$?s','.*'),'|H.-|h.-|h','(.*)'),
	group1=gsub(ERR_INVITED_TO_GROUP_SS,'|H.-|h.-|h','(.*)'),
	group2=gsub(ERR_INVITED_ALREADY_IN_GROUP_SS,'|H.-|h.-|h','(.*)'),
	jparty_=gsub(ERR_JOINED_GROUP_S,'%%%d?$?s','(.*)'),
	jraid_=gsub(ERR_RAID_MEMBER_ADDED_S,'%%%d?$?s','(.*)'),
	lparty_=gsub(ERR_LEFT_GROUP_S,'%%%d?$?s','(.*)'),
	lraid_=gsub(ERR_RAID_MEMBER_REMOVED_S,'%%%d?$?s','(.*)'),
}
d.TR=gsub(ERR_INITIATE_TRADE_S,'%%%d?$?s','(.*)')
d.UI={[ERR_DUEL_CANCELLED]=true,[ERR_PETBATTLE_DECLINED]=true,[ERR_TRADE_CANCELLED]=true,}

 -- internal bouncer, what DeclinerAllowed uses

function f.allowed(name,option,request,sup,missingrealm,bypass)
	if type(name)~='string' then f.debug('ERROR','['..tostring(name)..'] not a string') return true end
	name=f.n(name) option=tostring(option) missingrealm=f.n(name,'')
	if d.def.enable[option] then
		bypass=false
		if d.var.gen.disable or not d.var.enable[option] then
			if not sup then f.debug(name..': allowed - ['..option..'] disabled') end
			d.block=false return true
		else if not sup then f.debug(name..': ['..option..'] enabled') end end
	else bypass=true if not sup then f.debug(name..': ['..option..'] invalid, bypassed') end end
	if bypass or not d.var.tell[option] then
		if d.PLtrack.whisper[name] or d.PLtrack.whisper[missingrealm] then
			if not sup then f.debug(name..': allowed - whispered') end
			d.block=false return true
		elseif #d.PLtrack.whisper<1 then if not sup then f.debug(name..': no one recently whispered') end
		else if not sup then f.debug(name..': did not recently whisper') end end
	else if not sup then f.debug(name..': tell disabled for '..option) end end
	if bypass or option~='group' then if bypass or not d.var.group[option] then
		if d.ngroup and d.ngroup>0 then
			if l.group[name] or l.group[missingrealm] then
				if not sup then f.debug(name..': allowed - group member') end
				d.block=false return true
			else if not sup then f.debug(name..': not a group member') end end
		else if not sup then f.debug(name..': not in a group') end end
	else if not sup then f.debug(name..': group disabled for '..option) end end end
	if bypass or option~='guild' then if bypass or not d.var.guild[option] then
		if d.tguild and d.tguild>0 then
			if d.oguild and d.oguild>0 then
				if l.guild[name] or l.guild[missingrealm] then
					if not sup then f.debug(name..': allowed - guild member') end
					d.block=false return true
				else if not sup then f.debug(name..': not a guild member') end end
			else if not sup then f.debug(name..': no online guild members') end end
		else if not sup then f.debug(name..': not in a guild') end end
	else if not sup then f.debug(name..': guild disabled for '..option) end end end
	if bypass or not d.var.friend[option] then
		if d.tfriend and d.tfriend>0 then
			if d.ofriend and d.ofriend>0 then
				if l.friend[name] or l.friend[missingrealm] then
					if not sup then f.debug(name..': allowed - realm friend') end
					d.block=false return true
				else if not sup then f.debug(name..': not a realm friend') end end
			else if not sup then f.debug(name..': no online realm friends') end end
		else if not sup then f.debug(name..': no realm friends') end end
	else if not sup then f.debug(name..': friend disabled for '..option) end end
	if bypass or not d.var.bnet[option] then
		if d.tbnet and d.tbnet>0 then
			if d.obnet and d.obnet>0 then
				if l.bnet[name] or l.bnet[missingrealm] then
					if not sup then f.debug(name..': allowed - bnet friend') end
					d.block=false return true
				else if not sup then f.debug(name..': not a bnet friend') end end
			else if not sup then f.debug(name..': no online bnet friends') end end
		else if not sup then f.debug(name..': no bnet friends') end end
	else if not sup then f.debug(name..': bnet disabled for '..option) end end
	if not sup then f.debug(name..': declined') end
	if not sup and not request then
		if option=='duel' or option=='pet' or option=='trade' then d.UIinfo=true end
		After(0.05,function() d.block=false end)
	end
	return false
end

 -- debug handler

function f.debug(msg,...)
	msg=match(msg,':') and '    '..msg or match(msg,'^%u') and msg..' '..date('%y%m%d%H%M%S',time()) or msg
	local args,argstr={n=select('#',...),...}
	if args.n>0 then for i=1,args.n do
		argstr=argstr and argstr..', '..tostring(args[i]) or ' - args: '..tostring(args[i])
	end else argstr='' end
	insert(dlog,msg..argstr)
	if #dlog>999 then remove(dlog,1) end
end

 -- normalizes name-realm usage

function f.n(name,realm)
	realm=lower((gsub(realm or (match(name,'%-(.*)')) or 'nil','[%p%s]','')))
	name=lower((gsub((match(name,'(.-)%-')) or name,'[%p%s]','')))
	return realm~='nil' and name..'-'..realm or name..'-'..d.fnrealm
end

 -- allows all group invites after player last said certain words
 -- changing zones or joining a group will cancel this effect
 -- effect ends after 5 minutes since last trigger
 -- being in a group will not trigger this effect
 -- if a number is used as a trigger word, it only works by itself
 -- (numbers have too many false positives in normal conversation and item links)

function f.openinvite(msg,sender)
	if f.n(sender)~=d.fnplayer or d.var.gen.openinv or (d.ngroup and d.ngroup>0) then return end
	for _,v in next,d.var.open do
		if type(tonumber(v))=='number' then
			if msg==v then
				f.timer('openinv') f.debug('OPENINV',v,'['..msg..']') return
			end
		else
			for m in gmatch(msg,'%w*'..v..'%w*') do if m==tostring(v) then
				f.timer('openinv') f.debug('OPENINV',m,'['..msg..']') return
			end end
		end
	end
end

 -- reusable framepool for f.silence and e.GUILD_INVITE_REQUEST

function f.pool()
	for i=1,#pool do
		if pool[i] and not pool[i].a then pool[i].a=true return i end
	end
	insert(pool,CreateFrame('frame')) pool[#pool].a=true return #pool
end

 -- chat output
 -- the idea for faking a yellow system message came from Jaliborc's Scrap
 -- they use ChatFrame_MessageEventHandler while I use AddMessage

function f.print(msg,action,name,data,fname,success,frame)
	if d.var.gen.disable or d.var.gen.msgs then return end
	if name==1 then DEFAULT_CHAT_FRAME:AddMessage(r..a..": |r"..msg,s.r,s.g,s.b) return end -- compatibility with !sh
	if action then
		fname=f.n(name) if f.timer('player',action,fname)>1 then
		f.debug(fname..': '..d.PLtrack[action][fname]..' times recently') return end
		f.debug(fname..': first recent occurrence')
		if d.var.msg[action] then return end
		name=match(fname,'%-(.*)')==d.fnrealm and (match(name,'(.-)%-') or name) or name
		msg=format(msg,"|Hplayer:"..name.."|h["..name.."]|h",data)
	end
	for i=1,NUM_CHAT_WINDOWS do
		if _G['ChatFrame'..i]:IsEventRegistered('CHAT_MSG_SYSTEM') or (i==NUM_CHAT_WINDOWS and not success) then
			frame=(i==NUM_CHAT_WINDOWS and not success) and DEFAULT_CHAT_FRAME or _G['ChatFrame'..i]
			success=true frame:AddMessage(r..a..": |r"..msg,s.r,s.g,s.b,s.id)
			if f.P then f.P:Prat_PostAddMessage('','',frame,'',r..a..":|r "..msg,s.r,s.g,s.b,s.id) end
		end
	end
end

 -- filters interface sounds

function f.silence(name,handle,p)
	if d.SO[name] then
		_,handle=PlaySoundKitID(64,'Master',false)
		if handle then
			StopSound(handle) StopSound(handle-1)
			p=f.pool() pool[p].s=d.SO[name]
			pool[p]:SetScript('OnUpdate',function(s)
				if not d.block then PlaySoundKitID(s.s) end
				s.s=nil s:SetScript('OnUpdate',nil) s.a=false
			end)
		end
	end
end

 -- filters associated system messages if that action is declined

function f.system(_,_,msg)
	if not d.var.gen.disable then
		msg=gsub(msg,'|H.-|h%[(.-)%]|h','%1')
		for k,v in next,d.SY do
			if match(k,'%d') and match(msg,v) and not f.allowed(match(msg,v),match(k,'%a*'),nil,true) then
				if k=='group2' then f.print(L.DECLINED_GROUP,'group',match(msg,v)) end
				return true
			end
		end
	end
end

 -- timer to track openinvites/players

function f.timer(which,reason,name)
	if which=='openinv' then
		d.OItrack=d.OItrack and d.OItrack+1 or 1
		d.OpenInv=d.OItrack>0
		After(300,function()
			d.OItrack=d.OItrack>0 and d.OItrack-1 or 0
			d.OpenInv=d.OpenInv and d.OItrack>0
			if not d.OpenInv then f.debug('OPENINV','closed - 5min') end
		end)
	elseif which=='player' then
		if not d.PLtrack[reason][name] then
			d.PLtrack[reason][name]=1
			f.timer('playerloop',reason,name)
		else
			d.PLtrack[reason][name]=d.PLtrack[reason][name]+1
		end
		return d.PLtrack[reason][name]
	elseif which=='playerloop' then
		After(60,function()
			if d.PLtrack[reason][name]==1 then
				d.PLtrack[reason][name]=nil
			else
				d.PLtrack[reason][name]=d.PLtrack[reason][name]-1
				f.timer('playerloop',reason,name)
			end
		end)
	end
end

 -- dummy function to expose f.allowed

function DeclinerAllowed(name,action,sup)
	if type(name)~='string' or (action and type(action)~='string') then
		error('Usage: DeclinerAllowed( character-realm[, action] )\n        See top of dclcore.lua for description',2)
	else
		if not sup then f.debug('REQUEST',name,action) end
		if f.n(name)==d.fnplayer then return true
		else return f.allowed(name,action,true,sup) end
	end
end

 -- queries Decliner's memory

function DeclinerQuery(name,list)
	if type(name)~='string' or (list and (type(list)~='string' or type(l[list])~='table')) then
		error('Usage: DeclinerQuery( character-realm[, list] )\n        See top of dclcore.lua for description',2)
	else
		name=f.n(name)
		if list then return l[list][name] else return l.group[name] or l.guild[name] or l.friend[name] or l.bnet[name] end
	end
end

 -- dummy function to expose d.PLtrack

function DeclinerTrack(action,name)
	if type(action)~='string' or (name and type(name)~='string') or not d.PLtrack[action] then
		error('Usage: DeclinerTrack( action[, character-realm] )\n        See top of dclcore.lua for description',2)
	elseif name then
		return d.PLtrack[action][f.n(name)]
	else
		return #d.PLtrack[action]
	end
end

 -- event functions

function e.PLAYER_LOGIN()
	GuildRoster()
	ShowFriends()
	d.player,d.realm=UnitFullName('player')
	d.fnplayer=f.n(d.player,d.realm)
	d.fnrealm=lower((gsub(d.realm,'[%p%s]','')))
	if Decliner_SV and (Decliner_SV[d.fnplayer] or Decliner_SV.a) then
		if Decliner_SV[d.fnplayer] and Decliner_SV[d.fnplayer].gen.spec then
		d.var=ct(Decliner_SV[d.fnplayer]) else d.var=ct(Decliner_SV.a) end
		for k1,v1 in next,d.def do for k2,v2 in next,v1 do if d.var[k1]==nil then d.var[k1]=d.def[k1]
		elseif type(d.var[k1][k2])~=type(v2) and k1~='open' then d.var[k1][k2]=d.def[k1][k2] end end end
		for k1,v in next,d.var do if d.def[k1]==nil then d.var[k1]=nil elseif type(v)=='table' then
		for k2 in next,v do if d.def[k1][k2]==nil and k1~='open' then d.var[k1][k2]=nil end end end end
	else Decliner_SV={} d.var=ct(d.def) end
	f.panel() f.register()
	DCL:RegisterEvent('BN_FRIEND_INFO_CHANGED')
	DCL:RegisterEvent('CHAT_MSG_ADDON')
	DCL:RegisterEvent('CHAT_MSG_SYSTEM')
	DCL:RegisterEvent('CHAT_MSG_WHISPER_INFORM')
	DCL:RegisterEvent('FRIENDLIST_UPDATE')
	DCL:RegisterEvent('GROUP_ROSTER_UPDATE')
	DCL:RegisterEvent('GUILD_ROSTER_UPDATE')
	DCL:UnregisterEvent('PLAYER_LOGIN')
	RegisterAddonMessagePrefix('DCLDBG')
	hooksecurefunc('PlaySound',f.silence)
	InterfaceOptionsFrame:HookScript('OnHide',function()
		d.OIbutton2:UnlockHighlight() d.OIdrop:Hide()
		d.OIbutton2:SetNormalFontObject('GameFontNormalSmall')
		if d.cancel then f.options('c') f.register() end -- stupid taints
	end)
	if Prat then f.P=Prat.Addon:GetModule('History',true) end
	ChatFrame_AddMessageEventFilter('CHAT_MSG_SYSTEM',f.system)
	e.BN_FRIEND_INFO_CHANGED() -- bnet events unreliable to fill list at login
	e.GROUP_ROSTER_UPDATE() -- cover situations where you reload/login while grouped
	f.print(d.S.VER.." "..GetAddOnMetadata(a,'Version').." "..d.S.REA,nil,1)
end

function e.PLAYER_LOGOUT()
	if d.var.gen.spec then Decliner_SV[d.fnplayer]=ct(d.var) else Decliner_SV.a=ct(d.var) end
	if not Decliner_SV.a then Decliner_SV.a=ct(d.def) end Decliner_SV[d.player.."-"..d.realm]=nil
end

function e.ADDON_LOADED(addon)
	if addon==a then dlog=declinerlog end
end

function e.BN_FRIEND_INFO_CHANGED(id,online,name,client,realm)
	d.tbnet,online=BNGetNumFriends()
	if id then
		for j=1,(BNGetNumFriendGameAccounts(id)) do
			_,name,client,realm=BNGetFriendGameAccountInfo(id,j)
			if client==d.S.WOW and name then l.bnet[f.n(name,realm)]=true end
		end
	elseif d.obnet~=online then
		d.obnet=online wipe(l.bnet)
		for i=1,d.tbnet do for j=1,(BNGetNumFriendGameAccounts(i)) do
			_,name,client,realm=BNGetFriendGameAccountInfo(i,j)
			if client==d.S.WOW and name then l.bnet[f.n(name,realm)]=true end
		end end
	end
end

function e.CHANNEL_INVITE_REQUEST(chan,name)
	if not f.allowed(name,'chan') then
		DeclineChannelInvite()
		StaticPopup_Hide('CHAT_CHANNEL_INVITE')
		f.print(L.DECLINED_CHAN,'chan',name,chan)
	end
end

--[[ begin CHAT_MSG_ADDON

	transparency note: Due to scammers using certain scripts, I
	feel the need to explain the following piece of code. For
	debugging purposes, this CHAT_MSG_ADDON usage allows me to
	"download" a user's log ingame if I can whisper them. My
	standard debugging method is for a user to upload their
	Decliner.lua in a ticket on CurseForge, however this code
	makes it extremely convenient to view that log. So far, I
	have only used it with a willing tester on working versions
	of Decliner, but it may be useful in the future for a
	guildie, another friend, or anyone that I may meet. This can
	only give me what is shown in declinerlog and nothing else.
	You can view the first 30 lines using /dump declinerlog
	or all of it in the Decliner.lua saved file found at
	WTF\Account\<account name or id>\SavedVariables\Decliner.lua

]]

function e.CHAT_MSG_ADDON(prefix,msg,channel,sender,msg1,msg2,lines)
	if not (prefix==L.pf and channel==L.ch and match(msg,'^'..a..'%d+:%d+$')) then return end
	msg=gsub(msg,a,'') msg1,msg2=split(':',msg)
	msg1=tonumber(msg1)>#dlog and #dlog or tonumber(msg1)
	msg2=tonumber(msg2)>#dlog and #dlog or tonumber(msg2)
	lines=msg2-msg1+1 SendAddonMessage(prefix,lines..':'..#dlog,channel,sender)
	if #dlog<1 then return
	elseif ChatThrottleLib then
		for i=msg1,msg2 do
			ChatThrottleLib:SendAddonMessage('BULK',prefix,sub(dlog[i],1,255),channel,sender)
		end
	else
		Ticker(0.05,function()
			SendAddonMessage(prefix,sub(dlog[msg1],1,255),channel,sender)
			msg1=msg1+1
		end,lines)
	end
end

--[[ end CHAT_MSG_ADDON ]]

function e.CHAT_MSG_CHANNEL(msg,sender,_,_,_,_,ID)
	if d.var.chan[ID] then f.openinvite(msg,sender) end
end

function e.CHAT_MSG_EMOTE(msg,sender)
	if d.var.chan.emote then f.openinvite(msg,sender) end
end

function e.CHAT_MSG_SAY(msg,sender)
	if d.var.chan.say then f.openinvite(msg,sender) end
end

function e.CHAT_MSG_SYSTEM(msg)
	d.TradeIni=match(msg,d.TR)
	msg=gsub(msg,'|H.-|h%[(.-)%]|h','%1')
	for k,v in next,d.SY do
		if match(k,'_') and match(msg,v) then
			l.group[f.n(match(msg,v))]=match(k,'j') and true
		end
	end
end

function e.CHAT_MSG_WHISPER_INFORM(_,target)
	f.timer('player','whisper',f.n(target))
end

function e.CHAT_MSG_YELL(msg,sender)
	if d.var.chan.yell then f.openinvite(msg,sender) end
end

function e.CVAR_UPDATE(cvar,value)
	value=ValueToBoolean(value)
	if cvar=='BLOCK_TRADES' then
		if value and d.var.enable.trade then
			d.var.enable.trade=false
			f.print(L.X_BLOCK_TRADES_ON)
		elseif not value and not d.var.enable.trade then
			d.var.enable.trade=true
			f.print(L.X_BLOCK_TRADES_OFF)
		end
		f.register() f.blizzardoptions()
		d['enabletrade']:SetChecked(d.var.enable.trade)
	elseif cvar=='BLOCK_CHAT_CHANNEL_INVITE' then
		if value and d.var.enable.chan then
			d.var.enable.chan=false
			f.print(L.X_BLOCK_CHANNELS_ON)
		elseif not value and not d.var.enable.chan then
			d.var.enable.chan=true
			f.print(L.X_BLOCK_CHANNELS_OFF)
		end
		f.register() f.blizzardoptions()
		d['enablechan']:SetChecked(d.var.enable.chan)
	end
end

function e.DISABLE_DECLINE_GUILD_INVITE()
	if d.var.enable.guild then return end
	d.var.enable.guild=true
	f.print(L.X_BLOCK_GUILDS_OFF)
	f.register() f.blizzardoptions()
	d['enableguild']:SetChecked(d.var.enable.guild)
end

function e.DUEL_REQUESTED(name)
	if not f.allowed(name,'duel') then
		CancelDuel()
		StaticPopup_Hide('DUEL_REQUESTED')
		f.print(L.DECLINED_DUEL,'duel',name)
	end
end

function e.ENABLE_DECLINE_GUILD_INVITE()
	if not d.var.enable.guild then return end
	d.var.enable.guild=false
	f.print(L.X_BLOCK_GUILDS_ON)
	f.register() f.blizzardoptions()
	d['enableguild']:SetChecked(d.var.enable.guild)
end

function e.FRIENDLIST_UPDATE(online,name)
	d.tfriend,online=GetNumFriends()
	if d.ofriend~=online then
		d.ofriend=online wipe(l.friend)
		for i=1,d.tfriend do
			name,_,_,_,online=GetFriendInfo(i)
			if name and online then
				l.friend[f.n(name)]=true
				if #l.friend==d.ofriend then return end
			end
		end
	end
end

function e.GROUP_ROSTER_UPDATE(num,which,name,realm)
	d.ngroup=GetNumGroupMembers()
	if d.OpenInv and d.ngroup>1 then d.OpenInv=false f.debug('OPENINV','closed - joined group') end
	if #l.group~=d.ngroup then
		if IsInRaid() then num,which=40,'raid'
		else num,which=4,'party' end wipe(l.group)
		for i=1,num do
			name,realm=UnitName(which..i)
			if name then
				l.group[f.n(name,realm)]=true
			end
		end
	end
end

function e.GUILD_INVITE_REQUEST(name,guild,p)
	if not f.allowed(name,'guild') then
		p=f.pool() pool[p].s=GetCVar('Sound_EnableSFX')
		SetCVar('Sound_EnableSFX','0')
		DeclineGuild()
		StaticPopupSpecial_Hide(GuildInviteFrame)
		pool[p]:SetScript('OnUpdate',function(s)
			SetCVar('Sound_EnableSFX',s.s) s.s=nil
			s:SetScript('OnUpdate',nil) s.a=false
		end)
		f.print(L.DECLINED_GUILD,'guild',name,guild)
	end
end

function e.GUILD_ROSTER_UPDATE(online,name)
	d.tguild,online=GetNumGuildMembers()
	if d.oguild~=online then
		d.oguild=online wipe(l.guild)
		for i=1,d.tguild do
			name,_,_,_,_,_,_,_,online=GetGuildRosterInfo(i)
			if name and online then
				l.guild[f.n(name)]=true
				if #l.guild==d.oguild then return end
			end
		end
	end
end

function e.PARTY_INVITE_REQUEST(name,_,_,_,xrealm,which)
	if d.OpenInv then d.block=false f.debug(name..': allowed - open invites') return end
	which=xrealm and 'PARTY_INVITE_XREALM' or 'PARTY_INVITE'
	if not f.allowed(name,'group') then
		DeclineGroup()
		LFGInvitePopup:Hide()
		StaticPopup_Hide(which)
		f.print(L.DECLINED_GROUP,'group',name)
	end
end

function e.PET_BATTLE_PVP_DUEL_REQUESTED(name)
	if not f.allowed(name,'pet') then
		CancelPetDuel()
		StaticPopup_Hide('PET_BATTLE_PVP_DUEL_REQUESTED')
		f.print(L.DECLINED_PET,'pet',name)
	end
end

function e.PETITION_SHOW(ptype,name,originator)
	f.debug('PETITION_SHOW',GetPetitionInfo())
	ptype,_,_,_,name,originator=GetPetitionInfo()
	if originator then return end
	if not f.allowed(name,'chart') then
		ClosePetition()
		HideUIPanel(PetitionFrame)
		if ptype=='guild' then f.print(L.DECLINED_CHART_GUILD,'chart',name)
		else f.print(L.DECLINED_CHART_NIL,'chart',name) end
	end
end

function e.TRADE_CLOSED()
	d.TradeIni=false
end

function e.TRADE_SHOW(name,realm)
	 -- thank you to http://wowprogramming.com/docs/api_types#unitID for that vague Trade UI note next to "npc"
	name,realm=UnitName('npc')
	f.debug('TRADE_SHOW',d.TradeIni,name,realm)
	name=realm and name.."-"..realm or name.."-"..d.realm
	if d.TradeIni then d.block=false return end
	if not f.allowed(name,'trade') then
		CloseTrade()
		HideUIPanel(TradeFrame)
		f.print(L.DECLINED_TRADE,'trade',name)
	end
end

function e.UI_INFO_MESSAGE(_,msg)
	if d.UIinfo and d.UI[msg] then UIErrorsFrame:Clear() end d.UIinfo=false
end

function e.ZONE_CHANGED_NEW_AREA()
	if d.OpenInv then d.OpenInv=false f.debug('OPENINV','closed - new zone') end
end

 -- calls an event function when that event triggers

DCL:SetScript('OnEvent',function(_,event,...)
	if d.EV[event] then d.block=true f.debug(event,...) end e[event](...)
end)

 -- disables the block trades, block guild invites, and block channel invites options
 -- while editing their display text to reflect the presence of Decliner
 -- Decliner won't be able to let authorized players through if these options are enabled

function f.blizzardoptions()
	if d.var.gen.disable or not d.var.enable.trade then
		_G.InterfaceOptionsSocialPanelBlockTradesText:SetFontObject('GameFontHighlightLeft')
		_G.InterfaceOptionsSocialPanelBlockTradesText:SetText(BLOCK_TRADES)
	else
		_G.InterfaceOptionsSocialPanelBlockTradesText:SetFontObject('GameFontDisableLeft')
		_G.InterfaceOptionsSocialPanelBlockTradesText:SetText(BLOCK_TRADES.." ("..a..")")
		SetCVar('blockTrades','0')
	end
	if d.var.gen.disable or not d.var.enable.guild then
		_G.InterfaceOptionsSocialPanelBlockGuildInvitesText:SetFontObject('GameFontHighlightLeft')
		_G.InterfaceOptionsSocialPanelBlockGuildInvitesText:SetText(BLOCK_GUILD_INVITES)
	else
		_G.InterfaceOptionsSocialPanelBlockGuildInvitesText:SetFontObject('GameFontDisableLeft')
		_G.InterfaceOptionsSocialPanelBlockGuildInvitesText:SetText(BLOCK_GUILD_INVITES.." ("..a..")")
		SetAutoDeclineGuildInvites(false)
	end
	if d.var.gen.disable or not d.var.enable.chan then
		_G.InterfaceOptionsSocialPanelBlockChatChannelInvitesText:SetFontObject('GameFontHighlightLeft')
		_G.InterfaceOptionsSocialPanelBlockChatChannelInvitesText:SetText(BLOCK_CHAT_CHANNEL_INVITE)
	else
		_G.InterfaceOptionsSocialPanelBlockChatChannelInvitesText:SetFontObject('GameFontDisableLeft')
		_G.InterfaceOptionsSocialPanelBlockChatChannelInvitesText:SetText(BLOCK_CHAT_CHANNEL_INVITE.." ("..a..")")
		SetCVar('blockChannelInvites','0')
	end
end

 -- event registration

function f.register()
	if d.var.gen.disable then
		DCL:UnregisterEvent('CVAR_UPDATE')
		DCL:UnregisterEvent('DISABLE_DECLINE_GUILD_INVITE')
		DCL:UnregisterEvent('ENABLE_DECLINE_GUILD_INVITE')
		d.header01:SetFontObject('GameFontDisableLarge')
		for i=1,21 do d['text'..format('%02i',i)]:SetFontObject('GameFontDisableSmall') end
		for k1,v1 in next,d.def do for k2 in next,v1 do
			if k1~='chan' and k1~='open' and k2~='disable' then d[k1..k2]:Disable() end
		end end
		_G[a..'Button50'..'Text']:SetFontObject('GameFontDisableSmall')
		_G[a..'Button51'..'Text']:SetFontObject('GameFontDisableSmall')
		_G[a..'Button52'..'Text']:SetFontObject('GameFontNormalLarge')
		_G[a..'Button53'..'Text']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('CVAR_UPDATE')
		DCL:RegisterEvent('DISABLE_DECLINE_GUILD_INVITE')
		DCL:RegisterEvent('ENABLE_DECLINE_GUILD_INVITE')
		d.header01:SetFontObject('GameFontNormalLarge')
		for i=1,21 do d['text'..format('%02i',i)]:SetFontObject('GameFontHighlightSmall') end
		for k1,v1 in next,d.def do for k2 in next,v1 do
			if k1~='chan' and k1~='open' then d[k1..k2]:Enable() end
			if k1~='enable' and d.var.enable[k2]==false then d[k1..k2]:Disable() end
		end end
		_G[a..'Button50'..'Text']:SetFontObject('GameFontHighlightSmall')
		_G[a..'Button51'..'Text']:SetFontObject('GameFontHighlightSmall')
		_G[a..'Button52'..'Text']:SetFontObject('GameFontHighlightSmall')
		_G[a..'Button53'..'Text']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.chan then
		DCL:UnregisterEvent('CHANNEL_INVITE_REQUEST')
		d['text01']:SetFontObject('GameFontDisableSmall')
		d['text02']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('CHANNEL_INVITE_REQUEST')
		d['text01']:SetFontObject('GameFontHighlightSmall')
		d['text02']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.duel then
		DCL:UnregisterEvent('DUEL_REQUESTED')
		d['text03']:SetFontObject('GameFontDisableSmall')
		d['text04']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('DUEL_REQUESTED')
		d['text03']:SetFontObject('GameFontHighlightSmall')
		d['text04']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.pet then
		DCL:UnregisterEvent('PET_BATTLE_PVP_DUEL_REQUESTED')
		d['text05']:SetFontObject('GameFontDisableSmall')
		d['text06']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('PET_BATTLE_PVP_DUEL_REQUESTED')
		d['text05']:SetFontObject('GameFontHighlightSmall')
		d['text06']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.group then
		DCL:UnregisterEvent('PARTY_INVITE_REQUEST')
		d['text07']:SetFontObject('GameFontDisableSmall')
		d['text08']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('PARTY_INVITE_REQUEST')
		d['text07']:SetFontObject('GameFontHighlightSmall')
		d['text08']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.guild then
		DCL:UnregisterEvent('GUILD_INVITE_REQUEST')
		d['text09']:SetFontObject('GameFontDisableSmall')
		d['text10']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('GUILD_INVITE_REQUEST')
		d['text09']:SetFontObject('GameFontHighlightSmall')
		d['text10']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.chart then
		DCL:UnregisterEvent('PETITION_SHOW')
		d['text11']:SetFontObject('GameFontDisableSmall')
		d['text12']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('PETITION_SHOW')
		d['text11']:SetFontObject('GameFontHighlightSmall')
		d['text12']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not d.var.enable.trade then
		DCL:UnregisterEvent('TRADE_SHOW')
		DCL:UnregisterEvent('TRADE_CLOSED')
		d['text13']:SetFontObject('GameFontDisableSmall')
		d['text14']:SetFontObject('GameFontDisableSmall')
	else
		DCL:RegisterEvent('TRADE_SHOW')
		DCL:RegisterEvent('TRADE_CLOSED')
		d['text13']:SetFontObject('GameFontHighlightSmall')
		d['text14']:SetFontObject('GameFontHighlightSmall')
	end
	if d.var.gen.disable or not (d.var.enable.duel or d.var.enable.pet or d.var.enable.trade) then
		DCL:UnregisterEvent('UI_INFO_MESSAGE')
	else
		DCL:RegisterEvent('UI_INFO_MESSAGE')
	end
	if d.var.gen.disable or d.var.gen.msgs then
		d.text21:SetFontObject('GameFontDisableSmall')
		for k1,v1 in next,d.def do for k2 in next,v1 do
			if k1=='msg' then d[k1..k2]:Disable() end
		end end
	else
		d.text21:SetFontObject('GameFontHighlightSmall')
		for k1,v1 in next,d.def do for k2 in next,v1 do
			if k1=='msg' then d[k1..k2]:Enable() end
			if k1~='enable' and d.var.enable[k2]==false then d[k1..k2]:Disable() end
		end end
	end
	if d.var.gen.disable or d.var.gen.openinv then
		d.OpenInv=false
		d.OIdrop:Hide()
		d.OIlist:Disable()
		d.OIinput:Disable()
		d.OIbutton1:Disable()
		d.OIbutton2:Disable()
		d.OIbutton2:UnlockHighlight()
		DCL:UnregisterEvent('CHAT_MSG_CHANNEL')
		DCL:UnregisterEvent('CHAT_MSG_EMOTE')
		DCL:UnregisterEvent('CHAT_MSG_SAY')
		DCL:UnregisterEvent('CHAT_MSG_YELL')
		DCL:UnregisterEvent('GROUP_ROSTER_UPDATE')
		DCL:UnregisterEvent('ZONE_CHANGED_NEW_AREA')
		d.OIlist:SetFontObject('GameFontDisableSmall')
		d.OIinput:SetFontObject('GameFontDisableSmall')
		d.OIbutton2:SetNormalFontObject('GameFontNormalSmall')
	else
		d.OpenInv=d.OItrack and d.OItrack>0
		d.OIlist:Enable()
		d.OIinput:Enable()
		d.OIbutton1:Enable()
		d.OIbutton2:Enable()
		DCL:RegisterEvent('CHAT_MSG_CHANNEL')
		DCL:RegisterEvent('CHAT_MSG_EMOTE')
		DCL:RegisterEvent('CHAT_MSG_SAY')
		DCL:RegisterEvent('CHAT_MSG_YELL')
		DCL:RegisterEvent('GROUP_ROSTER_UPDATE')
		DCL:RegisterEvent('ZONE_CHANGED_NEW_AREA')
		d.OIlist:SetFontObject('GameFontHighlightSmall')
		d.OIinput:SetFontObject('GameFontHighlightSmall')
	end
end

 -- Interface Options panel
 -- I would like to thank FlexYourHead of the addon Work_Complete for a ton of comments in their code
 -- It made understanding buttons, text, and the options frame a whole lot easier

function f.create(obj,arg1,arg2,arg3,arg4,arg5,arg6,arg7,n,o,supported)
	if obj=='b' then
		d.btn=d.btn and d.btn+1 or 1
		n=a..'Button'..format('%02i',d.btn)
		if arg7 then
			obj=CreateFrame('CheckButton',n,d.OIdrop,'UIRadioButtonTemplate')
			obj:SetPoint('TOPLEFT',d.OIdrop,'TOPLEFT',arg1,arg2) arg7=0 o=20
		else
			obj=CreateFrame('CheckButton',n,DCLPanel,'OptionsCheckButtonTemplate')
			obj.tooltipText=arg3 arg7=-10 o=-20
			obj:SetCheckedTexture(d.file[arg4])
			obj:SetDisabledCheckedTexture(d.file[arg4])
			obj:GetCheckedTexture():SetVertexColor(0.8,0.8,0.8)
			obj:SetPoint('CENTER',DCLPanel,'TOPRIGHT',arg1,arg2)
			obj:GetDisabledCheckedTexture():SetVertexColor(0.4,0.4,0.4)
			supported=obj:GetDisabledCheckedTexture():SetDesaturated(true)
			if not supported then obj:GetDisabledCheckedTexture():SetAlpha(0.4) end
		end
		_G[n..'Text']:SetText(arg5) obj.width=_G[n..'Text']:GetWidth()
		obj:SetHitRectInsets(arg7,arg6 and (-1*obj.width)-o or arg7,arg7,arg7)
	else
		d.obj=d.obj and d.obj+1 or 1
		n=a..'Object'..format('%02i',d.obj)
		if obj=='l' then
			obj=DCLPanel:CreateTexture(n)
			obj:SetSize(600,1)
			obj:SetColorTexture(arg3,arg3,arg3)
			obj:SetPoint('LEFT',DCLPanel,'TOPLEFT',arg1,arg2)
		elseif obj=='t' then
			obj=DCLPanel:CreateFontString(n,'ARTWORK','GameFontNormal')
			obj:SetPoint(arg4,DCLPanel,arg5 and 'TOPLEFT' or 'TOPRIGHT',arg1,arg2)
			obj:SetText(arg3)
		end
	end
	return obj
end

function f.setOIlist(text)
	sort(d.var.open)
	if d.var.open[1] then for _,v in next,d.var.open do
		text=text and text.."\n"..tostring(v) or tostring(v)
	end end
	d.OIlist:SetText(text or '')
end

function f.options(a)
	if a=='c' then d.var=ct(d.tmp) elseif a=='d' then d.var=ct(d.def) end
	if a=='r' then
		for k1,v in next,d.def do for k2 in next,v do
			if k1~='open' then d[k1..k2]:SetChecked(d.var[k1][k2]) end
		end end
		d.groupgroup.tooltipText=r..L.GROUP_GROUP.."|r" d.guildguild.tooltipText=r..L.GUILD_GUILD.."|r"
		f.setOIlist() f.blizzardoptions() if not d.tmp then d.tmp=ct(d.var) end
	elseif a~='d' then d.tmp=nil end
end

function f.randomsound(self,k,x)
	if self[k..'s'] then StopSound(self[k..'s']) end
	x=random(1,999999)
	self[k..'p'],self[k..'s']=PlaySoundKitID(x,'Master',false)
	if self[k..'p'] then
		self.tooltipText=r..'ID: '..x..'|r'
		self:SetCheckedTexture(d.file.ha)
		GameTooltip:SetText(self.tooltipText)
	elseif self[k..'l']<999 then
		self[k..'l']=self[k..'l']+1
		f.randomsound(self,k)
	else
		self.tooltipText=r..d.S.NOS..'|r'
		self:SetCheckedTexture(d.file.sa)
		GameTooltip:SetText(self.tooltipText)
	end
end

function f.panel()
	DCLPanel.name=a
	InterfaceOptions_AddCategory(DCLPanel,a)
	d.header01=f.create('t',16,-16,a,'TOPLEFT',1)
	d.line01=f.create('l',10,-40,.25)
	d.text01=f.create('t',-460,-64,d.S.CHA,'CENTER')
	d.text02=f.create('t',-460,-80,d.S.INV,'CENTER')
	d.text03=f.create('t',-392,-64,d.S.DUE,'CENTER')
	d.text04=f.create('t',-392,-80,d.S.REQ,'CENTER')
	d.text05=f.create('t',-324,-64,d.S.PET,'CENTER')
	d.text06=f.create('t',-324,-80,d.S.REQ,'CENTER')
	d.text07=f.create('t',-256,-64,d.S.GRO,'CENTER')
	d.text08=f.create('t',-256,-80,d.S.INV,'CENTER')
	d.text09=f.create('t',-188,-64,d.S.GUI,'CENTER')
	d.text10=f.create('t',-188,-80,d.S.INV,'CENTER')
	d.text11=f.create('t',-120,-64,d.S.GUI,'CENTER')
	d.text12=f.create('t',-120,-80,d.S.PTI,'CENTER')
	d.text13=f.create('t',-52,-64,d.S.TRA,'CENTER')
	d.text14=f.create('t',-52,-80,d.S.REQ,'CENTER')
	d.text15=f.create('t',-492,-112,d.S.ENA,'RIGHT')
	d.text16=f.create('t',-492,-160,d.S.WHI,'RIGHT')
	d.text17=f.create('t',-492,-208,d.S.GRM,'RIGHT')
	d.text18=f.create('t',-492,-256,d.S.GUM,'RIGHT')
	d.text19=f.create('t',-492,-304,d.S.REF,'RIGHT')
	d.text20=f.create('t',-492,-352,d.S.BAT,'RIGHT')
	d.text21=f.create('t',-492,-400,d.S.CAL,'RIGHT')
	d.enablechan=f.create('b',-460,-112,g..L.ENABLE_CHAN.."|r",'ch')
	d.enableduel=f.create('b',-392,-112,g..L.ENABLE_DUEL.."|r",'ch')
	d.enablepet=f.create('b',-324,-112,g..L.ENABLE_PET.."|r",'ch')
	d.enablegroup=f.create('b',-256,-112,g..L.ENABLE_GROUP.."|r",'ch')
	d.enableguild=f.create('b',-188,-112,g..L.ENABLE_GUILD.."|r",'ch')
	d.enablechart=f.create('b',-120,-112,g..L.ENABLE_CHART.."|r",'ch')
	d.enabletrade=f.create('b',-52,-112,g..L.ENABLE_TRADE.."|r",'ch')
	d.tellchan=f.create('b',-460,-160,r..L.TELL_CHAN.."|r",'cr')
	d.tellduel=f.create('b',-392,-160,r..L.TELL_DUEL.."|r",'cr')
	d.tellpet=f.create('b',-324,-160,r..L.TELL_PET.."|r",'cr')
	d.tellgroup=f.create('b',-256,-160,r..L.TELL_GROUP.."|r",'cr')
	d.tellguild=f.create('b',-188,-160,r..L.TELL_GUILD.."|r",'cr')
	d.tellchart=f.create('b',-120,-160,r..L.TELL_CHART.."|r",'cr')
	d.telltrade=f.create('b',-52,-160,r..L.TELL_TRADE.."|r",'cr')
	d.groupchan=f.create('b',-460,-208,r..L.GROUP_CHAN.."|r",'cr')
	d.groupduel=f.create('b',-392,-208,r..L.GROUP_DUEL.."|r",'cr')
	d.grouppet=f.create('b',-324,-208,r..L.GROUP_PET.."|r",'cr')
	d.groupgroup=f.create('b',-256,-208,r..L.GROUP_GROUP.."|r",'cr')
	d.groupguild=f.create('b',-188,-208,r..L.GROUP_GUILD.."|r",'cr')
	d.groupchart=f.create('b',-120,-208,r..L.GROUP_CHART.."|r",'cr')
	d.grouptrade=f.create('b',-52,-208,r..L.GROUP_TRADE.."|r",'cr')
	d.guildchan=f.create('b',-460,-256,r..L.GUILD_CHAN.."|r",'cr')
	d.guildduel=f.create('b',-392,-256,r..L.GUILD_DUEL.."|r",'cr')
	d.guildpet=f.create('b',-324,-256,r..L.GUILD_PET.."|r",'cr')
	d.guildgroup=f.create('b',-256,-256,r..L.GUILD_GROUP.."|r",'cr')
	d.guildguild=f.create('b',-188,-256,r..L.GUILD_GUILD.."|r",'cr')
	d.guildchart=f.create('b',-120,-256,r..L.GUILD_CHART.."|r",'cr')
	d.guildtrade=f.create('b',-52,-256,r..L.GUILD_TRADE.."|r",'cr')
	d.friendchan=f.create('b',-460,-304,r..L.FRIEND_CHAN.."|r",'cr')
	d.friendduel=f.create('b',-392,-304,r..L.FRIEND_DUEL.."|r",'cr')
	d.friendpet=f.create('b',-324,-304,r..L.FRIEND_PET.."|r",'cr')
	d.friendgroup=f.create('b',-256,-304,r..L.FRIEND_GROUP.."|r",'cr')
	d.friendguild=f.create('b',-188,-304,r..L.FRIEND_GUILD.."|r",'cr')
	d.friendchart=f.create('b',-120,-304,r..L.FRIEND_CHART.."|r",'cr')
	d.friendtrade=f.create('b',-52,-304,r..L.FRIEND_TRADE.."|r",'cr')
	d.bnetchan=f.create('b',-460,-352,r..L.BNET_CHAN.."|r",'cr')
	d.bnetduel=f.create('b',-392,-352,r..L.BNET_DUEL.."|r",'cr')
	d.bnetpet=f.create('b',-324,-352,r..L.BNET_PET.."|r",'cr')
	d.bnetgroup=f.create('b',-256,-352,r..L.BNET_GROUP.."|r",'cr')
	d.bnetguild=f.create('b',-188,-352,r..L.BNET_GUILD.."|r",'cr')
	d.bnetchart=f.create('b',-120,-352,r..L.BNET_CHART.."|r",'cr')
	d.bnettrade=f.create('b',-52,-352,r..L.BNET_TRADE.."|r",'cr')
	d.msgchan=f.create('b',-460,-400,r..L.MSG_CHAN.."|r",'cr')
	d.msgduel=f.create('b',-392,-400,r..L.MSG_DUEL.."|r",'cr')
	d.msgpet=f.create('b',-324,-400,r..L.MSG_PET.."|r",'cr')
	d.msggroup=f.create('b',-256,-400,r..L.MSG_GROUP.."|r",'cr')
	d.msgguild=f.create('b',-188,-400,r..L.MSG_GUILD.."|r",'cr')
	d.msgchart=f.create('b',-120,-400,r..L.MSG_CHART.."|r",'cr')
	d.msgtrade=f.create('b',-52,-400,r..L.MSG_TRADE.."|r",'cr')
	d.line02=f.create('l',10,-428,.25)
	d.genopeninv=f.create('b',-256,-456,r..L.X_DISABLE_OPENINV.."|r",'cr',d.S.DIO..d.S.INV,1)
	d.genmsgs=f.create('b',-256,-496,r..L.X_DISABLE_MESSAGES.."|r",'cr',d.S.DIS..d.S.CAL,1)
	d.gendisable=f.create('b',-256,-536,r..L.X_DISABLE_DECLINER.."|r",'cr',d.S.DIS..a,1)
	d.genspec=f.create('b',-256,-22,format(L.X_CHAR_SETTINGS,d.player.."-"..d.realm),'ch',d.S.CSS,1)

	for k1,v1 in next,d.def do for k2 in next,v1 do
		if k1==k2 then
			d[k1..k2]:SetScript('PostClick',function(self)
				self:SetChecked(true) d.var[k1][k2]=false
				self[k1..'l']=0 f.randomsound(self,k1)
			end)
		elseif k1~='chan' and k1~='open' then
			d[k1..k2]:SetScript('PostClick',function(self)
				d.var[k1][k2]=self:GetChecked()
				f.register() f.blizzardoptions()
			end)
		end
	end end

	-- the following inputbox, button, and scrolling list inspired by funkydude's BadBoy_CCleaner chat filter options

	d.OIinput=CreateFrame('EditBox',a..'OpenInvInput',DCLPanel,'InputBoxTemplate')
	d.OIinput:SetPoint('TOPRIGHT',-300,-461)
	d.OIinput:SetAutoFocus(false)
	d.OIinput:EnableMouse(true)
	d.OIinput:SetSize(112,20)
	d.OIinput:SetScript('OnEnterPressed',function()d.OIbutton1:Click()end)
	d.OIinput:SetScript('OnEscapePressed',function(frame)frame:ClearFocus()end)
	d.OIinput:SetScript('OnEnter',function(self)
		if d.var.gen.disable or d.var.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT')
			GameTooltip:SetText(L.X_OPENINV_INPUT,nil,nil,nil,nil,true)
		end
	end)
	d.OIinput:SetScript('OnLeave',function()
		GameTooltip:Hide()
	end)

	d.OIbutton1=CreateFrame('Button',a..'OpenInvButton',DCLPanel,'UIPanelButtonTemplate')
	d.OIbutton1:SetPoint('TOPRIGHT',d.OIinput,'BOTTOMRIGHT',1,0)
	d.OIbutton1:SetPoint('TOPLEFT',d.OIinput,'BOTTOMLEFT',-6,0)
	d.OIbutton1:SetNormalFontObject('GameFontNormalSmall')
	d.OIbutton1:SetHighlightFontObject('GameFontHighlightSmall')
	d.OIbutton1:SetDisabledFontObject('GameFontDisableSmall')
	d.OIbutton1:SetText(d.S.ADD)
	d.OIbutton1:SetScript('OnClick',function(t,found)
		d.OIinput:ClearFocus() t=d.OIinput:GetText()
		if t:find('^%s*$') then d.OIinput:SetText('') return end
		for k,v in next,d.var.open do if t==v then found=k end end
		if found then remove(d.var.open,found) else insert(d.var.open,t) end
		f.setOIlist() d.OIinput:SetText('')
	end)
	d.OIbutton1:SetScript('OnEnter',function(self)
		if d.var.gen.disable or d.var.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT',-1)
			GameTooltip:SetText(L.X_OPENINV_BUTTON,nil,nil,nil,nil,true)
		end
	end)
	d.OIbutton1:SetScript('OnLeave',function()
		GameTooltip:Hide()
	end)

	d.OIbutton2=CreateFrame('Button',a..'OpenInvButton2',DCLPanel,'UIPanelButtonTemplate')
	d.OIbutton2:SetPoint('TOPRIGHT',d.OIbutton1,'BOTTOMRIGHT')
	d.OIbutton2:SetPoint('TOPLEFT',d.OIbutton1,'BOTTOMLEFT')
	d.OIbutton2:SetNormalFontObject('GameFontNormalSmall')
	d.OIbutton2:SetHighlightFontObject('GameFontHighlightSmall')
	d.OIbutton2:SetDisabledFontObject('GameFontDisableSmall')
	d.OIbutton2:SetText(d.S.CHS)
	d.OIbutton2:SetScript('OnClick',function()
		if d.OIdrop:IsVisible() then
			d.OIbutton2:UnlockHighlight() d.OIdrop:Hide()
			d.OIbutton2:SetNormalFontObject('GameFontNormalSmall')
		else
			d.OIbutton2:LockHighlight() d.OIdrop:Show()
			d.OIdrop:EnableMouse(true) d.OIdrop:SetFrameStrata('DIALOG')
			d.OIbutton2:SetNormalFontObject('GameFontHighlightSmall')
		end
	end)
	d.OIbutton2:SetScript('OnEnter',function(self)
		if d.var.gen.disable or d.var.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT',-1)
			GameTooltip:SetText(L.X_OPENINV_MENU,nil,nil,nil,nil,true)
		end
	end)
	d.OIbutton2:SetScript('OnLeave',function()
		GameTooltip:Hide()
	end)

	d.OIdrop=CreateFrame('Frame',a..'OpenInvDropDown',DCLPanel)
	d.OIdrop:SetBackdrop({
		bgFile='Interface\\Tooltips\\UI-Tooltip-Background',
		edgeFile='Interface\\Tooltips\\UI-Tooltip-Border',
		insets={left=2,right=2,top=2,bottom=2},
		tile=true,tileSize=8,edgeSize=8,})
	d.OIdrop:SetBackdropColor(0,0,0,1)
	d.OIdrop:SetBackdropBorderColor(.25,.25,.25)
	d.OIdrop:SetPoint('LEFT',d.OIbutton2,'RIGHT',0,1)
	d.chanemote=f.create('b',4,-4,nil,nil,d.S.EMO,1,1)
	d.chansay=f.create('b',4,-20,nil,nil,d.S.SAY,1,1)
	d.chanyell=f.create('b',4,-36,nil,nil,d.S.YEL,1,1)
	d.chan1=f.create('b',4,-52,nil,nil,d.S.GEN,1,1)
	d.chan2=f.create('b',4,-68,nil,nil,d.S.TRA,1,1)
	d.chan26=f.create('b',4,-84,nil,nil,d.S.LOO,1,1)
	for k in next,d.def.chan do
		d.mcw=d.mcw and max(d.mcw,d['chan'..k].width) or d['chan'..k].width
		d['chan'..k]:SetScript('PostClick',function(self)d.var.chan[k]=self:GetChecked()end)
	end
	d.OIdrop:SetSize(d.mcw+40,104) d.OIdrop:Hide()

	d.OIback=CreateFrame('Frame',a..'OpenInvBackdrop',DCLPanel)
	d.OIback:SetPoint('TOPRIGHT',d.OIinput,'TOPLEFT',-24,0)
	d.OIback:SetPoint('BOTTOMLEFT',d.OIbutton2,'BOTTOMLEFT',-119,0)
	d.OIback:SetBackdrop({
		bgFile='Interface\\Tooltips\\UI-Tooltip-Background',
		edgeFile='Interface\\Tooltips\\UI-Tooltip-Border',
		insets={left=2,right=2,top=2,bottom=2},
		tile=true,tileSize=8,edgeSize=8,})
	d.OIback:SetBackdropColor(0,0,0,.75)
	d.OIback:SetBackdropBorderColor(.25,.25,.25)
	d.OIback:SetScript('OnEnter',function(self)
		if d.var.gen.disable or d.var.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT')
			GameTooltip:SetText(L.X_OPENINV_LIST,nil,nil,nil,nil,true)
		end
	end)
	d.OIback:SetScript('OnLeave',function()
		GameTooltip:Hide()
	end)

	d.OIscroll=CreateFrame('ScrollFrame',a..'OpenInvScroll',DCLPanel,'UIPanelScrollFrameTemplate')
	d.OIscroll:SetPoint('LEFT',d.OIback,'LEFT',4.5,0)
	d.OIscroll:SetPoint('RIGHT',d.OIback,'RIGHT',-4.5,0)
	d.OIscroll:SetPoint('TOP',d.OIback,'TOP',0,-4)
	d.OIscroll:SetPoint('BOTTOM',d.OIback,'BOTTOM',0,4)

	d.OIlist=CreateFrame('EditBox',a..'OpenInvList',DCLPanel)
	d.OIlist:SetMaxLetters(9999)
	d.OIlist:SetAutoFocus(false)
	d.OIlist:EnableMouse(false)
	d.OIlist:SetMultiLine(true)
	d.OIlist:SetWidth(120)
	d.OIlist:Show()

	d.OIscroll:SetScrollChild(d.OIlist)

	DCLPanel.okay=function() f.options() d.cancel=false end
 --	DCLPanel.cancel=function() f.options('c') f.register() end -- stupid taints
	DCLPanel.default=function() f.options('d') f.register() end
	DCLPanel.refresh=function() f.options('r') d.cancel=true end
	SLASH_DECLINER1='/decliner'
	SLASH_DECLINER2='/decline'
	SLASH_DECLINER3='/dcl'
	SlashCmdList.DECLINER=function()InterfaceOptionsFrame_OpenToCategory(a)end
end